﻿//------------------------------------------------------------------------------------------------+
// Filename: KSLAM.cpp
// Content : SLAMを行う関数の定義など
//------------------------------------------------------------------------------------------------+

#include "KSLAM.h"
#include <cstdio>
#include <iostream>

#pragma warning ( disable : 4996 )

#define GLOBALMAP_WINDOW_NAME	"Localization Result"
#define GLOBALMAP_WINDOW_WIDTH	600
#define GLOBALMAP_WINDOW_HEIGHT	450
#define KAPPA	30.0					// 尤度計算の係数

using namespace std;


//------------------------------------------------------------------------------------------------+
// 初期化関数（コンストラクタ）
//------------------------------------------------------------------------------------------------+
KSLAM::KSLAM( int particle_num, SLAMinput *input, bool save_image, bool show_result, GlobalMap *g_map ){

	this->particle_num = particle_num;
	this->input        = input;
	this->show_result  = show_result;
	this->save_image   = save_image;

	// 画像パーツリスト用
	delete_P_count = 0;				// 地図パーツ再利用スタックポインタ
	Piece_num      = 0;				// 地図パーツの総数　Σ（各パーティクルが使っているパーツ数）

	Gmap_Piece* dummy;				// 先頭用ポインタ（基準用　使用はしない）
	dummy   = new Gmap_Piece();		// 先頭用ポインタ（基準用　使用はしない）
	dummy->set_keeping_num(10);		// 適当な閲覧数を不可（set_keeping_numは閲覧数のセットを行う）
	first_P = dummy;				// 先頭ポインタ初期化
	end_P   = dummy;				// 最後尾ポインタ初期化

	for( int i=0; i<REUSE_STACK_SIZE; i++ ) delete_P[i] = 0;

	// パーティクル用
	if( !g_map ) GrobalMap_L = new GlobalMap[particle_num];
	else GrobalMap_L = g_map;

	local_map = new LocalMap();

	sum_likelihood = 0;
	max_likelihood = -1e-200;
	resample.resize( particle_num );

	// 結果画像用
	GMAP = cvCreateImage( cvSize( GLOBALMAP_WINDOW_WIDTH, GLOBALMAP_WINDOW_HEIGHT ), IPL_DEPTH_8U, 3);

	world = input->RobotPose();		// 初期位置取得
	local_map->set_world( world );	// 局所地図の基準位置を登録

	old_world = world;

	first_flag = true;

	current_time = 1;

	if( show_result ) cvNamedWindow( GLOBALMAP_WINDOW_NAME );		// 結果表示用ウインドウ

	cerr << "Initialization of KSLAM is finished.\n";
}


//------------------------------------------------------------------------------------------------+
// デストラクタ
// メモリの開放をする
//------------------------------------------------------------------------------------------------+
KSLAM::~KSLAM(){

	if( show_result ) cvDestroyWindow( GLOBALMAP_WINDOW_NAME );

	Gmap_Piece* Pointer = first_P;

	while( Pointer != NULL ){

		Gmap_Piece* delete_Pi = Pointer;
		Pointer = Pointer->Next;	  
		delete delete_Pi;
	}

	delete [] GrobalMap_L;
	delete local_map;

	cvReleaseImage( &GMAP );
}


//------------------------------------------------------------------------------------------------+
// 地図生成を行う関数
//------------------------------------------------------------------------------------------------+
void KSLAM::Mapping(){

	if( !input->IsNewLRFData() || !input->IsNewPose() ) return;

	cerr << "\nStep: " << current_time << endl;

	IplImage *raw = 0;

	if( input->IsNewImage() && input->IsNewStereo() ) raw = input->Image();		// 画像の更新

	StereoData stereo = input->Stereo();		// ステレオ距離情報の更新
	world             = input->RobotPose();		// オドメトリ情報の更新
	RangeData range   = input->LRFdata();		// レーザセンサ情報の更新


	feature* input_features;
	int input_sift_num;
	vector<bool> effective_k_flag;
	vector<z_struct_3> z_data;

	// SIFT特徴を抽出
	if( raw ){
		input_sift_num   = sift_features( raw, &input_features );
		effective_k_flag = CheckEffective( input_sift_num, input_features );
		z_data           = MatchingLandmarkAndSIFT( input_sift_num, input_features, effective_k_flag, stereo );
	}

	// レーザー距離データを局所地図形式に変換
	local_map->input_data( range.max_data, range.max_angle, range.min_angle, range.value, MAX_RANGE );

	cerr << "Start to predict and to update Global Map\n";

	Predict( first_flag );

	if( raw ) UpdateGlobalMap( current_time, z_data );

	UpdateGlobalMap( current_time );

	cerr << "Finish to predict and to update Global Map\n";

	Likelihood( z_data );

	local_map->reset_map();		// 局所地図のリセット

	Resampling();
	UpdateMapPiece();

	if( show_result || save_image ){

		ShowImage( current_time, raw, input_sift_num, input_features, effective_k_flag, stereo, z_data );

		cvZero( GMAP );				// 画像のリセット
	}

	current_time++;
	first_flag = false;

	cerr << "Finish one cycle processing.\n\n";
}


//------------------------------------------------------------------------------------------------+
// グリッド地図・経路・ランドマーク地図の保存
//------------------------------------------------------------------------------------------------+
void KSLAM::SaveData(){

	SaveMapImage();
	SaveMapDataText();	
	SavePathData();
	SaveLandmarkMap();
}


//------------------------------------------------------------------------------------------------+
// 大域地図画像を返す（画像は呼び出した側で適切に開放されなければならない）
//------------------------------------------------------------------------------------------------+
IplImage *KSLAM::GlobalMapImage(){

	int best = GetBestGlobalMap();

	return GrobalMap_L[best].GlobalMapImage();
}


//------------------------------------------------------------------------------------------------+
// 大域地図の原点(0,0)の座標を返す
//------------------------------------------------------------------------------------------------+
Coordinate KSLAM::GlobalMapOrigin(){

	Coordinate origin;

	int best = GetBestGlobalMap();

	DataRange size = GrobalMap_L[best].GlobalMapSize();

	origin.x = -1.0 * G_grid_size / 1000.0 * G_MAP_size * ((double)( G_MAP_WIDTH  / 2 - size.min_x ) + 0.5 );
	origin.y = -1.0 * G_grid_size / 1000.0 * G_MAP_size * ((double)( G_MAP_HEIGHT / 2 - size.min_y ) + 0.5 );

	return origin;
}


//------------------------------------------------------------------------------------------------+
// 自己位置推定結果を返す
//------------------------------------------------------------------------------------------------+
Pose KSLAM::EstimatedPose(){

	int best = GetBestGlobalMap();

	int newest = GrobalMap_L[best].get_age() -1;

	return Pose( GrobalMap_L[best].path[newest].x, GrobalMap_L[best].path[newest].y, GrobalMap_L[best].path[newest].th );
}


//------------------------------------------------------------------------------------------------+
// 最も尤度の高いパーティクルを探す
// 返り値はGlobalMap（パーティクル）の添え字
//------------------------------------------------------------------------------------------------+
int KSLAM::GetBestGlobalMap(){

	int    best_likelihood       = 0;
	double best_likelihood_value = 0.0;

	for( int i=0; i<particle_num; i++ ){

		if( GrobalMap_L[i].get_sum_likelihood() < best_likelihood_value ){

			best_likelihood_value = GrobalMap_L[i].get_sum_likelihood();
			best_likelihood       = i;
		}
	}

	return best_likelihood;
}


//------------------------------------------------------------------------------------------------+
// グリッド地図を画像として保存する
//------------------------------------------------------------------------------------------------+
void KSLAM::SaveMapImage(){

	int best = GetBestGlobalMap();

	IplImage *global_map_image = cvCreateImage(cvSize ( LOCALMAP_WIDTH * 2, LOCALMAP_HEIGHT * 2 ), IPL_DEPTH_8U, 3 );

	GrobalMap_L[best].show_grid_map( global_map_image );

	// グリッド地図の保存
	cvSaveImage( "save/GmapA.png", global_map_image );
	cvZero( global_map_image );

	GrobalMap_L[best].clean_map((&end_P),delete_P,(&delete_P_count));
	GrobalMap_L[best].show_grid_map( global_map_image );

	cvSaveImage( "save/GmapB.png", global_map_image );

	cvReleaseImage( &global_map_image );
}


//------------------------------------------------------------------------------------------------+
// グリッド地図をテキストデータとして保存する
//------------------------------------------------------------------------------------------------+
void KSLAM::SaveMapDataText(){

	FILE *fl4;
	FILE *fl7;

	fl7 =  fopen( "save/map_likehood.txt", "wt" );

	for( int i=0; i<particle_num; i++ ){

		fprintf( fl7, "%d %lf %lf\n", i,GrobalMap_L[i].get_likelihood(), GrobalMap_L[i].get_sum_likelihood() );

		char name[1000];
		sprintf_s( name, "save/map%d.txt", i );
		fl4 = fopen( name, "wt" );
		GrobalMap_L[i].write_grid_map( fl4 );
		fclose( fl4 );
	}
	fclose( fl7 );
}


//------------------------------------------------------------------------------------------------+
// 経路をテキストデータとして保存する
//------------------------------------------------------------------------------------------------+
void KSLAM::SavePathData(){

	FILE *fl5;

	fl5 = fopen( "save/path.txt", "wt" );

	int best_likelihood = GetBestGlobalMap();

	for( int j=0; j<GrobalMap_L[best_likelihood].get_age(); j++ ){
		fprintf(fl5,"%d\t%lf\t%lf\t%lf\n",j,GrobalMap_L[best_likelihood].path[j].x
			,GrobalMap_L[best_likelihood].path[j].y
			,GrobalMap_L[best_likelihood].path[j].th);
	}

	fclose(fl5);
}


//------------------------------------------------------------------------------------------------+
// ランドマーク地図をテキストデータとして保存する
//------------------------------------------------------------------------------------------------+
void KSLAM::SaveLandmarkMap(){

	int three_LM_NUM = 0;

	int best_likelihood = GetBestGlobalMap();

	for( int i=0; i<(GrobalMap_L[best_likelihood].get_lmark_num()); i++ ){

		if( GrobalMap_L[best_likelihood].lmark[i].nsec>2 ) three_LM_NUM++;
	}


	// ランドマーク地図をテキストデータとして保存する
	FILE *fl3;
	char buf[1000];				//文字バッファ   
	sprintf_s( buf, "save/landmark.txt" ); 
	fl3 = fopen( buf, "wt" );

	fprintf( fl3, "%d\n", three_LM_NUM );

	for( int i=0;i<(GrobalMap_L[best_likelihood].get_lmark_num());i++){

		if( GrobalMap_L[best_likelihood].lmark[i].nsec>2 ){
			fprintf(fl3,"%lf\n",GrobalMap_L[best_likelihood].lmark[i].nsec);

			fprintf(fl3,"%lf\t%lf\t%lf\n",GrobalMap_L[best_likelihood].lmark[i].myu_x,
				GrobalMap_L[best_likelihood].lmark[i].myu_y,
				GrobalMap_L[best_likelihood].lmark[i].myu_z);

			for( int j=0; j<128; j++ ){
				fprintf(fl3,"%lf\n",GrobalMap_L[best_likelihood].lmark[i].SIFT_Val[j]);
			}

			for( int j=0; j<100; j++ ){
				fprintf(fl3,"%d\n",GrobalMap_L[best_likelihood].lmark[i].observation_time[j]);
			}

			for( int j=0; j<100; j++ ){
				fprintf(fl3,"%lf\n",GrobalMap_L[best_likelihood].lmark[i].discovery_degree_S[j]);
			}

			printf("%lf\n",GrobalMap_L[best_likelihood].lmark[i].discovery_degree);
			fprintf(fl3,"%lf\n",GrobalMap_L[best_likelihood].lmark[i].discovery_degree);
		}
	}
	fclose( fl3 );
}


//------------------------------------------------------------------------------------------------+
// 抽出されたSIFT特徴が有効かどうかのチェック
// trueなら条件に当てはまるSIFT特徴
//------------------------------------------------------------------------------------------------+
vector<bool> KSLAM::CheckEffective( int num, feature *features ){

	vector<bool> effective;

	for( int i=0; i<num; i++ ){

		// 同時刻の観測同士の近さを計算　一定値より近い観測がある場合使わない
		double min_C_E = -1.0;

		for(int j=0; j<num; j++) if( i != j ){

			double C_E = 0.0;

			for( int k=0; k<128; k++ ){

				//128次元の各特徴の差の和の2乗を計算(相違度)
				C_E += pow( features[i].descr[k] - features[j].descr[k], 2 );
			}

			// 一番近い時の相違度を計算
			if( min_C_E == -1.0 || min_C_E > C_E ) min_C_E = C_E;
		}

		//条件の確認
		if( sqrt( min_C_E ) < 250 && min_C_E != -1.0 ) effective.push_back( false );
		else effective.push_back( true );
	}

	return effective;
}


//------------------------------------------------------------------------------------------------+
// SIFTとランドマークの対応データを返す
//------------------------------------------------------------------------------------------------+
vector<z_struct_3> KSLAM::MatchingLandmarkAndSIFT( int num, feature *features, vector<bool> effective, StereoData stereo ){

	if( (int)effective.size() < num ) num = (int)effective.size();

	vector<z_struct_3> z_data;

	for( int i=0; i<num; i++ ) if( effective[i] ){

		//他に近い観測がないか

		int row = (int)( features[i].x );
		int col = (int)( features[i].y );

		// 位置条件　床と距離の制限
		if( 0.0 <= stereo.value[row][col][2]
			&& stereo.value[row][col][2] > 1500.0
			&& stereo.value[row][col][2] < 8000.0
			&& stereo.value[row][col][1] > -800.0
			&& stereo.value[row][col][1] < 800.0	   
			){

				z_struct_3 data;

				// 入力用変数にSIFTのデータを入力
				data.z_x = stereo.value[row][col][2];
				data.z_y = stereo.value[row][col][0];
				data.z_z = stereo.value[row][col][1];	
				data.window_x = row;
				data.window_y = col;
				data.match    = false;
				data.match_x  = 0.0;
				data.match_y  = 0.0;
				data.match_z  = 0.0;

				for( int xyz=0; xyz<128; xyz++ ) data.SIFT_Val[xyz] = features[i].descr[xyz];

				z_data.push_back( data );
		}
	}

	cerr << "Number of matched SIFT features = " << z_data.size() << endl;

	return z_data;
}


//------------------------------------------------------------------------------------------------+
// パーティクル（大域地図）の更新
//------------------------------------------------------------------------------------------------+
void KSLAM::Predict( bool first_loop ){

	double dis_theta = world.angle - old_world.angle;

	// パーティクルの分布範囲を決定する
	for( int p=0; p<particle_num; p++ ){

		double x_error  = 0.0;
		double y_error  = 0.0;
		double th_error = 0.0;

		// 初回の更新の場合はパーティクルを分散させない
		// （すべてのパーティクルを基準位置にするため）
		if( !first_loop ){

			x_error  = 100.0 * nrand();
			y_error  = 100.0 * nrand();
			th_error = 0.075 * nrand();
		}

		double moved_x = world.x - old_world.x + x_error;
		double moved_y = world.y - old_world.y + y_error;
/*		double value   = GrobalMap_L[p].GetCellValue( moved_x, moved_y );

		// 移動できないところに移動させない
		if( !first_loop ) for( double i=20.0; value < 0.0 || 0.3 <= value; i+=1.0 ){

			moved_x = world.x - old_world.x + i * nrand();
			moved_y = world.y - old_world.y + i * nrand();
			value   = GrobalMap_L[p].GetCellValue( moved_x, moved_y );
		}*/

		// 大域地図に移動量を入力
		GrobalMap_L[p].set_path( moved_x, moved_y, (dis_theta) + th_error, old_world.angle );
	}

	old_world = world;		// 現在の位置を保存
}


//------------------------------------------------------------------------------------------------+
// 各大域地図が持つランドマーク地図を更新
//------------------------------------------------------------------------------------------------+
void KSLAM::UpdateGlobalMap( int current_time, vector<z_struct_3> z_data ){

	for( int p=0; p<particle_num; p++ ) GrobalMap_L[p].input_z( z_data, current_time, 0 );
}


//------------------------------------------------------------------------------------------------+
// パーティクルの更新 （グリッド）
//------------------------------------------------------------------------------------------------+
void KSLAM::UpdateGlobalMap( int current_time ){

	for( int p=0; p<particle_num; p++ ){

		int aage   = GrobalMap_L[p].get_age();
		double dx  = GrobalMap_L[p].path[aage-1].x - GrobalMap_L[p].path[ aage - 1 - LMap_charge_time ].x;
		double dy  = GrobalMap_L[p].path[aage-1].y - GrobalMap_L[p].path[ aage - 1 - LMap_charge_time ].y;
		double dth = GrobalMap_L[p].path[aage-1].th - GrobalMap_L[p].path[ aage - 1 - LMap_charge_time ].th;

		GrobalMap_L[p].input_lmap( local_map, dx, dy, dth, GrobalMap_L[p].path[ aage - 1 - LMap_charge_time ].th, KAPPA, (&end_P), delete_P, (&delete_P_count) );
	}
}


//------------------------------------------------------------------------------------------------+
// パーテクルの尤度を計算する
//------------------------------------------------------------------------------------------------+
void KSLAM::Likelihood( vector<z_struct_3> z_data ){

	cerr << "Start to calculate Likelihoods\n";

	// 尤度の計算
	double  max_likelihood_L = DBL_MIN;
	double  max_likelihood_G = DBL_MIN;

	for(int p=0; p<particle_num; p++ ){
		if( max_likelihood_L < GrobalMap_L[p].get_likelihood_L() ) max_likelihood_L = GrobalMap_L[p].get_likelihood_L();
		if( max_likelihood_G < GrobalMap_L[p].get_likelihood_G() ) max_likelihood_G = GrobalMap_L[p].get_likelihood_G();
	}

	sum_likelihood = 0;		//selective resampleようにパーティクルすべての尤度の合計を記録する

	// SIFT特徴の数によって尤度を計算
	for( int p=0; p<particle_num; p++ ){

		//SIFTとグリッド両方使用
		if( 200 < z_data.size() ){

			GrobalMap_L[p].set_likelihood( GrobalMap_L[p].get_likelihood_L() / max_likelihood_L * GrobalMap_L[p].get_likelihood_G() / max_likelihood_G );
		}
		//グリッドの尤度だけ
		else GrobalMap_L[p].set_likelihood( GrobalMap_L[p].get_likelihood_G() / max_likelihood_G );

		sum_likelihood += GrobalMap_L[p].get_likelihood();	//合計尤度の計算
		GrobalMap_L[p].set_sum_likelihood( GrobalMap_L[p].get_sum_likelihood() * GrobalMap_L[p].get_likelihood() );		//累積尤度の登録
	}

	cerr << "Finish to calculate Likelihoods\n";
}


//------------------------------------------------------------------------------------------------+
// パーティクルのまき直し
//------------------------------------------------------------------------------------------------+
void KSLAM::Resampling(){

	// 有効なパーティクルの数を調べる
	double likelihood_var = 0;

	for( int i=0; i<particle_num; i++ ) likelihood_var += sq(GrobalMap_L[i].get_likelihood()/fabs(sum_likelihood));

	double N_eff = 1.0 / likelihood_var;

	if( N_eff < (double)particle_num / 2.0 ){ // 有効なパーティクルの数が半分以上ならresample

		cerr << "Start resampling process\n";
		vector<int> next_particle_num;
		next_particle_num.resize( particle_num );
		int flag2, p_num;

		for( int i=0; i<particle_num; i++ ){

			flag2 = 0;
			p_num = 0;
			next_particle_num[i] = 0;
			double r = ran0() * fabs(sum_likelihood);
			while( flag2 == 0 ){

				r -= GrobalMap_L[p_num].get_likelihood();
				if( r < 0 ) flag2 = 1;
				else p_num++;
			}
			resample[i] = p_num;
		}

		for( int i=0; i<particle_num; i++ ) next_particle_num[resample[i]]++;

		vector<int> delete_particle;
		delete_particle.resize( particle_num );
		int delete_particle_sum = 0;
		int copy_num = 0;

		for( int i=0; i<particle_num; i++ ){
			if( next_particle_num[i] == 0 ){
				delete_particle[delete_particle_sum] = i;
				delete_particle_sum++;
			}
			// printf("i=%d  num=%d   like=%lf\n",i,next_particle_num[i],GrobalMap_L[i].get_likelihood());
		}
		delete_particle_sum--;

		for( int i=0; i<particle_num; i++ ){
			copy_num = next_particle_num[i];

			while( copy_num > 1 ){
				GrobalMap_L[delete_particle[delete_particle_sum]].copy_data(GrobalMap_L[i]);
				delete_particle_sum--;
				copy_num--;
			} 
		}


		// 尤度のリセット
		for( int i=0; i<particle_num; i++ ){ 
			GrobalMap_L[i].set_likelihood( 1.0 );
			GrobalMap_L[i].set_sum_likelihood( 1.0 );
			GrobalMap_L[i].set_likelihood_L( 1.0 );
			GrobalMap_L[i].set_likelihood_G( 1.0 );
		}

		cerr << "Finish resampling process\n";
	}
}


//------------------------------------------------------------------------------------------------+
// 地図パーツのリストの更新
// 閲覧数0のパーツをリストから除外
//------------------------------------------------------------------------------------------------+
void KSLAM::UpdateMapPiece(){

	Gmap_Piece* Pointer = first_P; 
	Piece_num = 0;

	while( Pointer!=NULL ){//リストの最後尾までたどる
		if( Pointer->keeping_num == 0 ){//閲覧数が0のパーツのとき
			Pointer->Before->Next=Pointer->Next;//ポインタをつなぎかえる（一回前パーツと次のパーツのポインタをつなぐ）
			if( Pointer->Next == NULL ){//リストが最後尾の時
				end_P = Pointer->Before;
			}
			else{
				Pointer->Next->Before=Pointer->Before;//ポインタをつなぎかえる（一回前パーツと次のパーツのポインタをつなぐ）
			}	  
			if( delete_P_count < 10000 ){		//再利用のスタックの容量が開いてるときスタックに廃棄パーツのポインタをいれる
				delete_P[delete_P_count] = Pointer;
				Pointer = (Pointer->Next);
				delete_P[delete_P_count]->Next = NULL;
				delete_P[delete_P_count]->Before = NULL;
				delete_P_count++;
			}
			else{//スタックがいっぱいの時は直接パーツを消す
				printf("----------------OVER=============================");
				Gmap_Piece* delete_P2 = NULL;
				delete_P2 = Pointer;
				Pointer = (Pointer->Next);
				if( delete_P2 != NULL ){
					delete delete_P2;
					delete_P2=NULL;
				}
			}
		}
		else{
			Piece_num++;//有効パーツ数のカウント
			Pointer = ( Pointer->Next );
		}
	}
}


//------------------------------------------------------------------------------------------------+
// 表示用の画像を生成する
//------------------------------------------------------------------------------------------------+
void KSLAM::ShowImage( int current_time, IplImage *raw, int num, feature *features, vector<bool> effective, StereoData stereo, vector<z_struct_3> z_data ){

	if( raw ){

		double resize = IMG_WIDTH / raw->width;

		// 画像が大きいので半分サイズにして表示
		IplImage *raw_half = cvCreateImage( cvSize(IMG_WIDTH/2,IMG_HEIGHT/2 ), raw->depth, raw->nChannels );
		cvResize( raw, raw_half, CV_INTER_NN );
		show_Image_add( GMAP, raw_half, 0, 0 );
		cvReleaseImage( &raw_half );

		// ステレオ距離画像の生成と表示
		// 画像が大きいので半分サイズに変換する
		IplImage *stereo_img = MakeStereoImage( stereo );
		IplImage *stereo_half = cvCreateImage( cvSize( IMG_WIDTH/2, IMG_HEIGHT/2 ), IPL_DEPTH_8U, 3 );
		cvResize( stereo_img, stereo_half, CV_INTER_NN );
		show_Image_add( GMAP, stereo_half, 0, 240 );
		cvReleaseImage( &stereo_img );
		cvReleaseImage( &stereo_half );

		for( int i=0; i<num; i++ ){

			// SIFT特徴の表示
			CvPoint pt1;
			pt1.x = (int)( resize * features[i].x / 2.0 );		// 位置計算
			pt1.y = (int)( resize * features[i].y / 2.0 );
			CvScalar rcolor = CV_RGB( 0, 0, 255 );
			cvCircle( GMAP, pt1, 1, rcolor );
			
			if( effective[i] ){

				int row = (int)( features[i].x );
				int col = (int)( features[i].y );

				// 位置条件　床と距離の制限
				if( 0.0 <= stereo.value[row][col][2]
					&& stereo.value[row][col][2] > 1500.0
					&& stereo.value[row][col][2] < 8000.0
					&& stereo.value[row][col][1] > -800.0
					&& stereo.value[row][col][1] < 800.0	   
					){

						// 条件をクリアしたSIFT特徴を赤い点で表示
						CvPoint pt1;
						pt1.x = (int)( resize * (double)row / 2.0 );
						pt1.y = (int)( resize * (double)col / 2.0 );
						CvScalar rcolor = CV_RGB (255,0,0);
						cvCircle( GMAP, pt1, 1,rcolor );
					}
			}
		}
	}

	int best_likelihood = GetBestGlobalMap();
	GrobalMap_L[best_likelihood].show_grid_map( GMAP );				// グリッド地図
	GrobalMap_L[best_likelihood].show_path( GMAP, 155, 100, 0 );	// 経路
	if( raw ) GrobalMap_L[best_likelihood].show_Imark( GMAP, z_data, IMG_WIDTH / raw->width );		// ランドマーク

//	for( int i=0; i<particle_num; i++ ) GrobalMap_L[i].show_pose( GMAP, 255, 0, 0 );	// パーティクルの分散
	for( int i=0; i<particle_num; i++ ){
		
		if( GrobalMap_L[i].GetCellValue() < 0.0 ) GrobalMap_L[i].show_pose( GMAP, 0, 255, 0 );
		else if( GrobalMap_L[i].GetCellValue() < 0.3 ) GrobalMap_L[i].show_pose( GMAP, 255, 0, 0 );
		else GrobalMap_L[i].show_pose( GMAP, 0, 255, 255 );
	}

	if( show_result ) cvShowImage ( GLOBALMAP_WINDOW_NAME, GMAP );
	cvWaitKey(5);

	// 結果画像の保存
	if( save_image ){
		char save_name[1000];
		sprintf_s( save_name, "gmap/gmapB%06d.jpg", current_time );
		cvSaveImage( save_name, GMAP );
	}
}


//------------------------------------------------------------------------------------------------+
// ステレオ距離画像の作成
//------------------------------------------------------------------------------------------------+
IplImage *KSLAM::MakeStereoImage( StereoData stereo ){

	cerr << "Generate Stereo Depth Image\n";
	IplImage *dImage_hsv = cvCreateImage( cvSize( stereo.image_width, stereo.image_height ), IPL_DEPTH_8U, 3 );
	IplImage *stereo_img = cvCreateImage( cvSize( stereo.image_width, stereo.image_height ), IPL_DEPTH_8U, 3 );

	for( int y=0; y<stereo.image_height; y++ ) for( int x=0; x<stereo.image_width; x++ ){

		if( stereo.value[x][y][2] > 0 ){
			Iblue(  dImage_hsv, x, y ) = min( (int)( 15.0 * stereo.value[x][y][2]/1000.0 ), 120 );
			Igreen( dImage_hsv, x, y ) = Ired( dImage_hsv, x, y ) = 255;
		}
		else Igreen( dImage_hsv, x, y ) = Ired( dImage_hsv, x, y ) = 0;
	}

	cvCvtColor( dImage_hsv, stereo_img, CV_HSV2BGR );

	cvReleaseImage( &dImage_hsv );

	return stereo_img;
}
